﻿using AutoMapper;
using Microsoft.AspNetCore.Authorization;
using Microsoft.AspNetCore.Mvc;
using SegMyBill.Web.Application.Service.Contracts;
using SegMyBill.Web.Distributed.DTO.DTOModels;
using SegMyBill.Web.Domain.Entities;
using System;
using System.Collections.Generic;
using System.Linq;
using FirebaseAdmin.Auth;
using System.Threading.Tasks;
using System.Net;
using Newtonsoft.Json.Linq;
using System.Net.Http.Headers;
using System.Text;
using System.Net.Http;
using Microsoft.Extensions.Configuration;
using Microsoft.AspNetCore.Http;
using System.Threading;
using System.IO;
using SegMyBill.Web.CrossCutting.ResponseSerializer;

namespace SegMyBill.Web.Controllers
{
    public class TemplateController : Controller
    {
        private readonly IMapper _mapper;
        private readonly Serilog.ILogger _logger;
        private readonly ITemplateService _templateService;
        private readonly IConfiguration Configuration;

        public TemplateController(IMapper mapper, Serilog.ILogger logger, ITemplateService templateService, IConfiguration configuration)
        {
            _mapper = mapper;
            _logger = logger;
            _templateService = templateService;
            Configuration = configuration;
        }

        [AllowAnonymous]
        [HttpGet("/templates")]
        public ActionResult<IEnumerable<TemplateDTO>> GetTemplates()
        {
            try
            {
                _logger.Information("Getting all the Templates");
                var resultMapped = _mapper.Map<IEnumerable<TemplateDTO>>(_templateService.GetAll().Result);

                return Ok(resultMapped);

            }
            catch (Exception ex)
            {
                _logger.Error(ex.Message);
                return StatusCode((int)HttpStatusCode.NotFound, ResponseSerializer.ResponseToJson((int)HttpStatusCode.NotFound, ex.Message));
            }
            return StatusCode((int)HttpStatusCode.NotFound, ResponseSerializer.ResponseToJson((int)HttpStatusCode.NotFound, "There are not templates"));
        }

        [Authorize]
        [HttpPost("/templates/new")]
        public async Task<ActionResult<string>> CreateTemplate([FromForm] IFormFile TemplateValue, [FromHeader] string authorization)
        {
            if (AuthenticationHeaderValue.TryParse(authorization, out var headerValue))
            {
                var token = await FirebaseAuth.DefaultInstance.VerifyIdTokenAsync(headerValue.Parameter);

                var creatorName = token.Uid;

                if (TemplateValue.FileName.EndsWith(".yml"))
                {
                    var stream = new MemoryStream();
                    TemplateValue.CopyTo(stream);

                    var templateStream = stream.ToArray();
                    string base64 = Convert.ToBase64String(templateStream);

                    _logger.Information("Creating template with name: " + TemplateValue.FileName);

                    var templateMetadata = new TemplateDTO { CreatorEmail = creatorName, DateOfCreation = DateTime.Now.ToString("dd/MM/yyyy  HH:mm:ss"), TemplateName = TemplateValue.FileName.Remove(TemplateValue.FileName.Length - 4), Value = base64 };
                    var mappedTemplate = _mapper.Map<TemplateEntity>(templateMetadata);
                    var succes = _templateService.createTemplate(mappedTemplate);
                    if (succes.Result)
                    {
                        return Ok(ResponseSerializer.ResponseToJson((int)HttpStatusCode.OK, "Succes"));
                    }
                    else
                    {
                        _logger.Error("Error creating Template with name: " + TemplateValue.FileName);
                        return StatusCode((int)HttpStatusCode.InternalServerError, ResponseSerializer.ResponseToJson((int)HttpStatusCode.InternalServerError, "Error creating the template."));
                    }
                }
                else
                {
                    return StatusCode((int)HttpStatusCode.BadRequest, ResponseSerializer.ResponseToJson((int)HttpStatusCode.BadRequest, "The template must be a .yml"));
                }
                
            }
            return StatusCode((int)HttpStatusCode.NonAuthoritativeInformation, ResponseSerializer.ResponseToJson((int)HttpStatusCode.NonAuthoritativeInformation, "Not authentificated correctly."));
        }

        [Authorize]
        [HttpDelete("/templates/delete")]
        public async Task<ActionResult<string>> DeleteTemplate(string name, [FromHeader] string authorization)
        {
            if (AuthenticationHeaderValue.TryParse(authorization, out var headerValue))
            {
                var token = await FirebaseAuth.DefaultInstance.VerifyIdTokenAsync(headerValue.Parameter);

                var isAdmin = (bool)token.Claims.FirstOrDefault(x => x.Key == "Admin").Value;

                if (isAdmin)
                {
                    var succes = _templateService.deleteTemplate(name);
                    if (succes.Result)
                    {
                        return Ok(ResponseSerializer.ResponseToJson((int)HttpStatusCode.OK, "Succes"));
                    }
                    else
                    {
                        return StatusCode((int)HttpStatusCode.BadRequest, ResponseSerializer.ResponseToJson((int)HttpStatusCode.BadRequest, "Template not found."));
                    }
                }
                return StatusCode((int)HttpStatusCode.NonAuthoritativeInformation, ResponseSerializer.ResponseToJson((int)HttpStatusCode.NonAuthoritativeInformation, "Not authorized."));
            }

            return StatusCode((int)HttpStatusCode.NonAuthoritativeInformation, ResponseSerializer.ResponseToJson((int)HttpStatusCode.NonAuthoritativeInformation, "Not authentificated correctly."));

        }


        [AllowAnonymous]
        [HttpGet("/templates/getByName")]
        public async Task<ActionResult<TemplateDTO>> GetTemplateByName(string name)
        {
            var resultMapped = _mapper.Map<TemplateDTO>(_templateService.GetTemplateByName(name).Result);
            if (resultMapped != null)
            {
                return Ok(resultMapped);
            }
            else
            {
                return StatusCode((int)HttpStatusCode.NoContent, "0 templates with this name.");
            }
        }

        [Authorize]
        [HttpPost("/getSegmentatedBill")]
        public async Task<ContentResult> GetSegmentatedBill([FromForm] IFormFile Pdf, [FromForm] string Template)
        {
            
            try
            {
                var httpClient = new HttpClient();

                Stream pdfStream = Pdf.OpenReadStream();
                MultipartFormDataContent content = new MultipartFormDataContent();
                content.Add(new StreamContent(pdfStream), "Pdf", Pdf.FileName);
                content.Add(new StringContent(Template), "Template");

                var response = await httpClient.PostAsync("http://localhost:4000/getJsonB64Template", content);
                
                var byteArray = await response.Content.ReadAsByteArrayAsync();
                var result = Encoding.UTF8.GetString(byteArray, 0, byteArray.Length);

                var returnValue = Content(result, "application/json");
                returnValue.StatusCode = 200;

                return returnValue;

            }
            catch(Exception ex)
            {
                _logger.Error(ex.Message);
            }

            var returnValue1 = Content(ResponseSerializer.ResponseToJson((int)HttpStatusCode.InternalServerError, "Internal Error, try later..."));
            returnValue1.StatusCode = 500;
            return returnValue1;
        }
        

        /*
        [AllowAnonymous]
        [HttpGet("/templates/getYmlByName")]
        public async Task<ActionResult<File>> GetTemplateYmlByName(string name)
        {
            var base64Yml = _mapper.Map<TemplateDTO>(_templateService.GetTemplateByName(name).Result).Value;

            Byte[] bytes = Convert.FromBase64String(base64Yml);
            System.Text.Encoding.UTF8.GetString(bytes);

            if (resultMapped != null)
            {
                return Ok(resultMapped);
            }
            else
            {
                return StatusCode((int)HttpStatusCode.NoContent, "0 templates with this name.");
            }
        }*/



    }
}
