﻿using AutoMapper;
using Microsoft.AspNetCore.Authorization;
using Microsoft.AspNetCore.Mvc;
using SegMyBill.Web.Application.Service.Contracts;
using SegMyBill.Web.Distributed.DTO.DTOModels;
using SegMyBill.Web.Domain.Entities;
using System;
using System.Collections.Generic;
using System.Linq;
using FirebaseAdmin.Auth;
using System.Threading.Tasks;
using System.Net;
using Newtonsoft.Json;
using Newtonsoft.Json.Linq;
using System.Net.Http.Headers;
using System.Text;
using System.Net.Http;
using Microsoft.Extensions.Configuration;
using SegMyBill.Web.CrossCutting.ResponseSerializer;

namespace SegMyBill.Web.Controllers
{
    public class UserController : Controller
    {
        private readonly IMapper _mapper;
        private readonly Serilog.ILogger _logger;
        private readonly IUserService _userService;
        private readonly IConfiguration Configuration;

        public UserController(IMapper mapper, Serilog.ILogger logger, IUserService userService, IConfiguration configuration)
        {
            _mapper = mapper;
            _logger = logger;
            _userService = userService;
            Configuration = configuration;
        }

        [Authorize]
        [HttpGet("/users")]
        public async Task<ActionResult<IEnumerable<UserDTO>>> GetUsers([FromHeader] string authorization)
        {
            if (AuthenticationHeaderValue.TryParse(authorization, out var headerValue))
            {
                var token = await FirebaseAuth.DefaultInstance.VerifyIdTokenAsync(headerValue.Parameter);

                var isAdmin = (bool)token.Claims.FirstOrDefault(x => x.Key == "Admin").Value;

                if (isAdmin)
                {
                    try
                    {
                        _logger.Information("Getting all the Users");
                        var resultMapped = _mapper.Map<IEnumerable<UserDTO>>(_userService.GetAll().Result);

                        return Ok(resultMapped);

                    }
                    catch (Exception ex)
                    {
                        _logger.Error(ex.Message);
                    }
                    return StatusCode((int)HttpStatusCode.NotFound, "Content not found");
                }
                return StatusCode((int)HttpStatusCode.NonAuthoritativeInformation, "Not authorized.");
            }

            return StatusCode((int)HttpStatusCode.NonAuthoritativeInformation, "Not authentificated correctly.");
        }


        [Authorize]
        [HttpPost("/users/new")]
        public async Task<ActionResult<string>> CreateUser([FromBody] UserDTO user, [FromHeader] string authorization)
        {
            if (AuthenticationHeaderValue.TryParse(authorization, out var headerValue))
            {
                var token = await FirebaseAuth.DefaultInstance.VerifyIdTokenAsync(headerValue.Parameter);

                var isAdmin = (bool)token.Claims.FirstOrDefault(x => x.Key == "Admin").Value;

                if (isAdmin)
                {
                    _logger.Information("Creating user with email: " + user.Email);
                    var mappedUser = _mapper.Map<UserEntity>(user);
                    var succes = _userService.createUser(mappedUser);
                    if (succes.Result)
                    {
                        return Ok(ResponseSerializer.ResponseToJson((int)HttpStatusCode.OK, "Succes"));
                    }
                    else
                    {
                        _logger.Error("Error creating user with email: " + user.Email);
                        return NotFound(ResponseSerializer.ResponseToJson((int)HttpStatusCode.NotFound, "Error creating user"));
                    }
                }
                return StatusCode((int)HttpStatusCode.NonAuthoritativeInformation, ResponseSerializer.ResponseToJson((int)HttpStatusCode.NonAuthoritativeInformation, "Not authorized."));
            }

            return StatusCode((int)HttpStatusCode.NonAuthoritativeInformation, ResponseSerializer.ResponseToJson((int)HttpStatusCode.NonAuthoritativeInformation, "Not authentificated correctly."));
        }


        [Authorize]
        [HttpDelete("/users/delete")]
        public async Task<ActionResult<string>> DeleteUser(string email, [FromHeader] string authorization)
        {
            if (AuthenticationHeaderValue.TryParse(authorization, out var headerValue))
            {
                var token = await FirebaseAuth.DefaultInstance.VerifyIdTokenAsync(headerValue.Parameter);

                var isAdmin = (bool)token.Claims.FirstOrDefault(x => x.Key == "Admin").Value;

                if (isAdmin)
                {
                    var succes = _userService.deleteUser(email);
                    if (succes.Result)
                    {
                        return Ok(ResponseSerializer.ResponseToJson((int)HttpStatusCode.OK, "Succes"));
                    }
                    else
                    {
                        return NotFound(ResponseSerializer.ResponseToJson((int)HttpStatusCode.NotFound,"User not found"));
                    }
                }
                return StatusCode((int)HttpStatusCode.NonAuthoritativeInformation, ResponseSerializer.ResponseToJson((int)HttpStatusCode.NonAuthoritativeInformation, "Not authorized."));
            }

            return StatusCode((int)HttpStatusCode.NonAuthoritativeInformation, ResponseSerializer.ResponseToJson((int)HttpStatusCode.NonAuthoritativeInformation, "Not authentificated correctly."));
        }


        [Authorize]
        [HttpGet("/users/getByEmail")]
        public async Task<ActionResult<UserDTO>> GetUserByEmail(string email, [FromHeader] string authorization)
        {
            if (AuthenticationHeaderValue.TryParse(authorization, out var headerValue))
            {
                var token = await FirebaseAuth.DefaultInstance.VerifyIdTokenAsync(headerValue.Parameter);

                var isAdmin = (bool)token.Claims.FirstOrDefault(x => x.Key == "Admin").Value;

                if (isAdmin)
                {
                    var resultMapped = _mapper.Map<UserDTO>(_userService.findByEmail(email).Result);
                    if (resultMapped != null)
                    {
                        return Ok(resultMapped);
                    }
                    else
                    {
                        return StatusCode((int)HttpStatusCode.NoContent, "0 users with this email."); //TODO FER COSETES
                    }

                }
                return StatusCode((int)HttpStatusCode.NonAuthoritativeInformation, "Not authorized.");
            }

            return StatusCode((int)HttpStatusCode.NonAuthoritativeInformation, "Not authentificated correctly.");
        }

        [AllowAnonymous]
        [HttpPost("/SignUp")]
        public async Task<ActionResult<string>> SignUpUserAsync([FromBody] UserDTO user)
        {
            try
            {
                var claims = new Dictionary<string, object>(){
                    { "Admin", false },
                };

                var a = await FirebaseAuth.DefaultInstance.CreateUserAsync(new UserRecordArgs
                {
                    Uid = user.Email,
                    Email = user.Email,
                    Password = user.Password
                });
                await FirebaseAuth.DefaultInstance.SetCustomUserClaimsAsync(a.Uid, claims);

                var json = JsonConvert.SerializeObject(new GoogleTokenRequestDTO { email = user.Email, password = user.Password, returnSecureToken = true });
                var data = new StringContent(json, Encoding.UTF8, "application/json");
                var url = Configuration["GoogleTokenApiURL"];
                using var client = new HttpClient();
                var response = await client.PostAsync(url, data);
                var result = await response.Content.ReadAsStringAsync();

                var jsonResult = (JObject)JsonConvert.DeserializeObject(result);

                var anyError = jsonResult.Properties().Any(x => x.Name.StartsWith("error"));

                if (anyError)
                {
                    int code = jsonResult["error"]["code"].Value<int>();
                    string errorMessage = jsonResult["error"]["message"].Value<string>();
                    return StatusCode((int)HttpStatusCode.BadRequest, ResponseSerializer.ResponseToJson((int)HttpStatusCode.BadRequest, errorMessage));
                }
                else
                {
                    _logger.Information("SignUp: Creating user with email: " + user.Email);
                    user.Role = 1;
                    var mappedUser = _mapper.Map<UserEntity>(user);
                    var succes = _userService.createUser(mappedUser);
                    if (succes.Result)
                    {
                        var token = jsonResult["idToken"].Value<string>();
                        _logger.Information("SignUp: Succes user with email: " + user.Email);

                        return Ok(ResponseSerializer.ResponseToJson((int)HttpStatusCode.OK,token));
                    }
                }

            }
            catch (Exception ex)
            {
                _logger.Error("Error in SignUp: " + ex.Message);
                if (ex.Source == "FirebaseAdmin")
                {
                    return StatusCode((int)HttpStatusCode.BadRequest, ResponseSerializer.ResponseToJson((int)HttpStatusCode.BadRequest, ex.Message));
                }
            }
            return StatusCode((int)HttpStatusCode.InternalServerError, ResponseSerializer.ResponseToJson((int)HttpStatusCode.InternalServerError, "Failed doing Sign Up"));
        }

        [AllowAnonymous]
        [HttpPost("/LogIn")]
        public async Task<ActionResult<string>> LogInUserAsync([FromBody]LoginRequestDTO loginRequest)
        {
            try
            {
                var json = JsonConvert.SerializeObject(new GoogleTokenRequestDTO { email = loginRequest.Email, password = loginRequest.Password, returnSecureToken = true });
                var data = new StringContent(json, Encoding.UTF8, "application/json");
                var url = Configuration["GoogleTokenApiURL"];
                using var client = new HttpClient();
                var response = await client.PostAsync(url, data);
                var result = await response.Content.ReadAsStringAsync();

                var jsonResult = (JObject)JsonConvert.DeserializeObject(result);

                var anyError = jsonResult.Properties().Any(x=>x.Name.StartsWith("error"));

                if (anyError)
                {
                    int code = jsonResult["error"]["code"].Value<int>();
                    string errorMessage = jsonResult["error"]["message"].Value<string>();
                    return StatusCode((int)HttpStatusCode.BadRequest, ResponseSerializer.ResponseToJson((int)HttpStatusCode.BadRequest,errorMessage));
                }
                else
                {
                    var token = jsonResult["idToken"].Value<string>();
                    _logger.Information("Login: Creating user with email: " + loginRequest.Email);

                    return Ok(ResponseSerializer.ResponseToJson((int)HttpStatusCode.OK, token));
                }
                
            }
            catch (Exception ex)
            {
                _logger.Error("Error in login: " + ex.Message);
            }
            return StatusCode((int)HttpStatusCode.InternalServerError, ResponseSerializer.ResponseToJson((int)HttpStatusCode.InternalServerError, "Failed doing the Login"));
        }


        [Authorize]
        [HttpPost("/SignUpWithRoles")]
        public async Task<ActionResult<string>> SignUpUserWithRoleAsync([FromBody] UserDTO user, [FromHeader] string authorization)
        {
            if (AuthenticationHeaderValue.TryParse(authorization, out var headerValue))
            {
                var token = await FirebaseAuth.DefaultInstance.VerifyIdTokenAsync(headerValue.Parameter);
                var isAdmin = (bool)token.Claims.FirstOrDefault(x => x.Key == "Admin").Value;

                if (isAdmin)
                {
                    try
                    {
                        var claims = new Dictionary<string, object>(){
                            { "Admin", user.Role==0 },
                        };

                        var a = await FirebaseAuth.DefaultInstance.CreateUserAsync(new UserRecordArgs
                        {
                            Uid = user.Email,
                            Email = user.Email,
                            Password = user.Password
                        });
                        await FirebaseAuth.DefaultInstance.SetCustomUserClaimsAsync(a.Uid, claims);

                        var json = JsonConvert.SerializeObject(new GoogleTokenRequestDTO { email = user.Email, password = user.Password, returnSecureToken = true });
                        var data = new StringContent(json, Encoding.UTF8, "application/json");
                        var url = Configuration["GoogleTokenApiURL"];
                        using var client = new HttpClient();
                        var response = await client.PostAsync(url, data);
                        var result = await response.Content.ReadAsStringAsync();
                        
                        var jsonResult = (JObject)JsonConvert.DeserializeObject(result);

                        var anyError = jsonResult.Properties().Any(x => x.Name.StartsWith("error"));

                        if (anyError)
                        {
                            int code = jsonResult["error"]["code"].Value<int>();
                            string errorMessage = jsonResult["error"]["message"].Value<string>();
                            return StatusCode((int)HttpStatusCode.BadRequest, ResponseSerializer.ResponseToJson((int)HttpStatusCode.BadRequest, errorMessage));
                        }
                        else
                        {
                            _logger.Information("SignUp: Creating user with email: " + user.Email);
                            var mappedUser = _mapper.Map<UserEntity>(user);
                            var succes = _userService.createUser(mappedUser);
                            if (succes.Result)
                            {
                                var tokenresult = jsonResult["idToken"].Value<string>();
                                _logger.Information("SignUp with Roles: Succes user with email: " + user.Email);
                                return Ok(ResponseSerializer.ResponseToJson((int)HttpStatusCode.OK, tokenresult));
                            }
                        }
                        
                    }
                    catch (Exception ex)
                    {
                        _logger.Error("Error in SignUp: " + ex.Message, ex.StackTrace);
                        if(ex.Source == "FirebaseAdmin")
                        {
                            return StatusCode((int)HttpStatusCode.BadRequest, ResponseSerializer.ResponseToJson((int)HttpStatusCode.BadRequest, ex.Message));
                        }

                    }
                    return StatusCode((int)HttpStatusCode.InternalServerError, ResponseSerializer.ResponseToJson((int)HttpStatusCode.InternalServerError, "Failed doing Sign Up"));
                }
                return StatusCode((int)HttpStatusCode.NonAuthoritativeInformation, ResponseSerializer.ResponseToJson((int)HttpStatusCode.NonAuthoritativeInformation, "Not authorized."));
            }
            return StatusCode((int)HttpStatusCode.NonAuthoritativeInformation, ResponseSerializer.ResponseToJson((int)HttpStatusCode.NonAuthoritativeInformation, "Not authentificated correctly."));
        }


        [Authorize]
        [HttpDelete("/DeleteUserAuth")]
        public async Task<ActionResult<bool>> deleteUserAuthAsync(string email, [FromHeader] string authorization)
        {
            if (AuthenticationHeaderValue.TryParse(authorization, out var headerValue))
            {
                var token = await FirebaseAuth.DefaultInstance.VerifyIdTokenAsync(headerValue.Parameter);
                var isAdmin = (bool)token.Claims.FirstOrDefault(x => x.Key == "Admin").Value;

                if (isAdmin)
                {
                    var result = true;
                    try
                    {
                        await FirebaseAuth.DefaultInstance.DeleteUserAsync(email);
                        result = _userService.deleteUser(email).Result;

                    }
                    catch (Exception ex)
                    {
                        result = false;
                        _logger.Error("Error deleting UserAuth: " + ex.Message, ex.StackTrace);
                        return StatusCode((int)HttpStatusCode.InternalServerError, ResponseSerializer.ResponseToJson((int)HttpStatusCode.InternalServerError, "Failed deleting UserAuth"));
                    }
                    return Ok(ResponseSerializer.ResponseToJson((int)HttpStatusCode.InternalServerError, "Succes"));
                }
                return StatusCode((int)HttpStatusCode.NonAuthoritativeInformation, ResponseSerializer.ResponseToJson((int)HttpStatusCode.NonAuthoritativeInformation, "Not authorized."));
            }
            return StatusCode((int)HttpStatusCode.NonAuthoritativeInformation, ResponseSerializer.ResponseToJson((int)HttpStatusCode.NonAuthoritativeInformation, "Not authentificated correctly."));
        }

        [Authorize]
        [HttpGet("/UserRoleByToken")]
        public async Task<ActionResult<string>> getUserRoleByToken([FromHeader] string authorization)
        {
            if (AuthenticationHeaderValue.TryParse(authorization, out var headerValue))
            {
                var token = await FirebaseAuth.DefaultInstance.VerifyIdTokenAsync(headerValue.Parameter);
                var isAdmin = (bool)token.Claims.FirstOrDefault(x => x.Key == "Admin").Value;

                if (isAdmin)
                {
                    return Ok(ResponseSerializer.ResponseToJson((int)HttpStatusCode.OK, "Admin"));
                }
                return Ok(ResponseSerializer.ResponseToJson((int)HttpStatusCode.OK, "Normal"));
                
            }

            return StatusCode((int)HttpStatusCode.NonAuthoritativeInformation, "Not authentificated correctly.");
        }

    }

}
