from . import resolvers
from .dataStructures import lines, tables
from collections import OrderedDict
import dateparser
import re

RESOLVER_SELECTOR = {"lines": resolvers.lines, "regex": resolvers.regex}
DATA_STRUCTURE_SELECTOR = {"lines": lines, "tables": tables}

DEFAULT_TEMPLATE_SETTINGS = {
    "date_formats": [],
    "languages": [],
    "currency": "EUR",
    "decimal_separator": ",",
}


class BillTemplate(OrderedDict):
    # Aquesta classe és la representació de una plantilla tipus .yml

    def __init__(self, *args, **kwargs):
        super(BillTemplate, self).__init__(*args, **kwargs)

        # Ajunta les opcions de la plantilla amb les per dfefecte
        self.options = DEFAULT_TEMPLATE_SETTINGS.copy()

        # Mira si tots els llenguatges que s'han posat a la plantilla estan formats per dues lletres
        for language in self.options["languages"]:
            assert len(language) == 2, "language code must have 2 letters"

        # Sobreescriu les opcions segons el contingut de la template
        if "template_settings" in self:
            self.options.update(self["template_settings"])

        # En cas que no tingui el camp empresa agafa el primer valor dels camps obligatoris
        if "business" not in self.keys():
            self["business"] = self["must_to_words"][0]

    def clean_number(self, value):
        # Aquesta funció s'ocupa de treure els separadors d'una xifra i assignar-li el separador decimal ".", per
        # facilitar string->float.

        # Comprova que no hi hagi més d'un separador decimal dins d'un valor
        assert (
                value.count(self.options["decimal_separator"]) < 2
        ), "No es pot declarar més d'un separador decimal a la vegada"

        # Modifica el separador decimal per #
        decimalMarkTmpReplace = value.replace(self.options["decimal_separator"], "#")
        # a través del patró r"[.,'\s]", substituit-lo per un lloc en blanc es pot eliminar tots els separadors de
        # milers
        cleanedSeparators = re.sub(r"[.,'\s]", "", decimalMarkTmpReplace)
        # per defecte es modifica al separador decimal anglés
        return float(cleanedSeparators.replace("#", "."))

    def matches_all_mtw(self, cleaned_string: str) -> bool:
        # Fa la comprovació de que totes les must_to_words de la plantilla estiguin dins del text

        if all([re.search(keyword, cleaned_string) for keyword in self["must_to_words"]]):
            return True
        return False

    def parse_date_by_format(self, date):
        # Fa el parse de la data a partir dels valors que s'han donat a les opcions de la template.

        dateParsed = dateparser.parse(
            date,
            date_formats=self.options["date_formats"],
            languages=self.options["languages"],
        )
        return dateParsed

    def to_type(self, value, trans_type):
        # Modifica els tipus dels valors I ORDRE

        if trans_type == "float":
            if not value.strip():
                return 0.0
            else:
                return float(self.clean_number(value))
        elif trans_type == "int":
            if not value.strip():
                return 0
            else:
                return int(self.clean_number(value))
        elif trans_type == "date":
            return self.parse_date_by_format(value)
        assert False, "Tipus desconegut!"

    def extract(self, bill_string):
        # A partir del text del pdf, extreu les dades que conté

        # Com que l'empresa serà introduida com un valor estatic i string no hem de fer res més
        extraction = {"business": self["business"]}

        # field_name = el seguent variable dels camps principals de la template
        # field_value = el contingut d'aquella variable de camp de la template
        for field_name, field_value in self["basics"].items():
            if field_name.startswith("static_"):
                # sempre que el camp comença amb static_ significa que el contingut està fixat en la mateixa plantilla
                extraction[field_name.replace("static_", "")] = field_value
            else:
                content = None
                if field_name.startswith("date") or field_name.endswith("date"):
                    content = resolvers.regex.parse(self, field_name, {"regex": field_value, "type": "date"},
                                                    bill_string, True)
                elif field_name.startswith("amount"):
                    content = resolvers.regex.parse(self, field_name, {"regex": field_value, "type": "float"},
                                                    bill_string, True)
                else:
                    content = resolvers.regex.parse(self, field_name, {"regex": field_value}, bill_string, True)

                if content is not None:
                    # Asigna a la sortida el resultat que ha donat el match
                    extraction[field_name] = content

        extraction["currency"] = self.options["currency"]  # Assigna la moneda posada a les opcions

        # extreu les funcions i a on son localitzades les funcions de serialització de taules i lineas
        for function_name, function_location in DATA_STRUCTURE_SELECTOR.items():
            if function_name in self.keys():  # si un dels plugins esta contingut a un dels camps de la template:
                # a la localització del plugin executa el seu metode d'extracció
                function_location.pull_out(self, bill_string, extraction)

        # Per ser una factura si o si ha de contenir aquests camps!
        if "must_to_fields" not in self.keys():
            must_to_fields = ["date", "amount", "invoice_number", "business"]
        else:
            must_to_fields = []
            for field_value in self["must_to_fields"]:
                must_to_fields.append(field_value)

        # els must_to_fields formen part de lo extret de la plantilla?
        if set(must_to_fields).issubset(extraction.keys()):
            extraction["description"] = "Bill from %s" % (self["business"])
            return extraction
        else:
            # TODO THROW EXCEPTION: FALTEN CAMPS OBLIGATORIS O NO TROBA ELS SEUS PATRONS
            return None
